/**
 * An extended JSON library that supports the encoding and decoding of arbitrary
 * objects.
 */
library jsonx;

import 'dart:mirrors';
import 'dart:convert' show JSON, JsonEncoder;

/**
 * Decodes the JSON string [text].
 *
 * The optional [reviver] function is called once for each object or list
 * property that has been parsed during decoding. The `key` argument is either
 * the integer list index for a list property, the map string for object
 * properties, or `null` for the final result.
 *
 * The default [reviver] (when not provided) is the identity function.
 *
 * The optional [type] parameter specifies the type to which [text] should be
 * decoded. Since Dart doesn't allow passing a generic type as an argument, one must
 * create an instance of that generic type and pass the instance's runtimeType
 * as the value of [type].
 *
 * If [type] is omitted, this method is equivalent to [JSON.decode] in
 * **dart:convert** library.
 *
 * Example:
 *
 *     class Person {
 *       String name;
 *       int age;
 *     }
 *
 *     Person p = decode('{ "name": "Man", "age": 20 }', Person);
 *     print(p.name);
 *
 *     List<int> list = decode('[1,2,3]', <int>[].runtimeType);
 */
// TODO: consider caching mirrors on [type].
decode(String text, {reviver(key, value), Type type}) {
  var json = JSON.decode(text, reviver: reviver);
  if (type == null) return json;
  return _jsonToObject(json, reflectType(type));
}

/**
 * Encodes [object] as a JSON string.
 *
 * The encoding happens as below:
 * 1. Tries to encode [object] directly
 * 2. If (1) fails, tries to call [object.toJson()] to convert [object] into
 * an encodable value
 * 3. If (2) fails, tries to use mirrors to convert [object] into en encodable
 * value
 *
 * Example:
 *
 *     class Person {
 *       Person(this.name, this.age);
 *       String name;
 *       int age;
 *     }
 *
 *     var p = new Person('kin', 20);
 *     print(encode(p));
 */
String encode(object) => _ENCODER.convert(object);

const _EMTPY_SYMBOL = const Symbol('');

_jsonToObject(json, mirror) {
  if (_isPrimitive(json)) return json;

  if (json is List) {
    var result = new List(json.length);
    var typeMirror = mirror.typeArguments.single;
    for (var i = 0; i < json.length; i++) {
      result[i] = _jsonToObject(json[i], typeMirror);
    }
    return result;
  }

  var instanceMirror = mirror.newInstance(_EMTPY_SYMBOL, []);

  if (instanceMirror.reflectee is Map) {
    var result = {};
    var typeMirror = mirror.typeArguments.last;
    for (var key in json.keys) {
      result[key] = _jsonToObject(json[key], typeMirror);
    }
    return result;
  }

  for (var key in json.keys) {
    var name = new Symbol(key);
    var decl = mirror.declarations[name];
    var type;

    if (_isPublicVariable(decl)) {
      type = decl.type;
    } else {
      var n = new Symbol('$key=');
      decl = mirror.declarations[n];
      if (_isPublicSetter(decl)) {
        type = decl.parameters.first.type;
      } else {
        continue;
      }
    }
    instanceMirror.setField(name, _jsonToObject(json[key], type));
  }
  return instanceMirror.reflectee;
}

bool _isPublicVariable(DeclarationMirror v) {
  return v is VariableMirror &&
      !v.isStatic && !v.isPrivate &&
      !v.isFinal;
  // TODO: Add isConst when the bug is fixed.
}

bool _isPublicSetter(DeclarationMirror v) {
  return v is MethodMirror && !v.isStatic && v.isSetter;
}

bool _isPublicGetter(DeclarationMirror v) {
  if (v.isPrivate) return false;
  if (v is VariableMirror && !v.isStatic) return true;
  if (v is MethodMirror && !v.isStatic && v.isGetter) return true;
  return false;
}

bool _isPrimitive(v) {
  return v is num || v is bool || v is String || v == null;
}

const _ENCODER = const JsonEncoder(_toEncodable);

_toEncodable(object) {
  try {
    return object.toJson();
  } catch (_) {
    return __toEncodable(object);
  }
}

__toEncodable(object) {
  if (_isPrimitive(object)) return object;

  if (object is List) {
    var list = [];
    for (var e in object) {
      list.add(__toEncodable(e));
    }
    return list;
  }

  var map = {};

  if (object is Map) {
    for (var key in object.keys) {
      map[key] = __toEncodable(object[key]);
    }
    return map;
  }

  var instanceMirror = reflect(object);

  instanceMirror.type.declarations.forEach((field, mirror) {
    if (_isPublicGetter(mirror)) {
      var name = MirrorSystem.getName(field);
      var value = instanceMirror.getField(field).reflectee;
      map[name] = __toEncodable(value);
    }
  });

  return map;
}